import * as React from 'react';
import {AugmentedFile, augmentFile, ContentType, getMimeTypeDescription} from '@/utils/blob';
import { FilePreview as FilePreviewComponent} from '../FilePreview';
import {formatFileSize} from '@/utils/numeral';
import {AudioMiniFilePreview} from '@tesseract-design/web-blob-react';
import {delegateTriggerChangeEvent} from '@/utils/event';

export interface FileButtonProps extends Omit<React.HTMLProps<HTMLInputElement>, 'size' | 'type' | 'style' | 'label' | 'list'> {
  /**
   * Should the component display a border?
   */
  border?: boolean,
  /**
   * Should the component occupy the whole width of its parent?
   */
  block?: boolean,
  /**
   * Short textual description indicating the nature of the component's value.
   */
  label?: React.ReactNode,
  /**
   * Short textual description as guidelines for valid input values.
   */
  hint?: React.ReactNode,
  enhanced?: boolean,
  /**
   * Is the label hidden?
   */
  hiddenLabel?: boolean,
}

const useFilePreviews = (fileList?: FileList) => {
  const [selectedFiles, setSelectedFiles] = React.useState([] as AugmentedFile[]);
  React.useEffect(() => {
    const loadFilePreviews = async (fileList: FileList) => {
      const files = Array.from(fileList);
      return Promise.all(
        files.map(async (f) => {
          const augmentedFile = await augmentFile(f);
          if (augmentedFile.resolvedType === ContentType.TEXT && augmentedFile.metadata?.scheme) {
            await import(`prismjs/components/prism-${augmentedFile.metadata.scheme}`);
          }
          return augmentedFile;
        })
      );
    }

    if (fileList) {
      loadFilePreviews(fileList).then((fileResult) => {
        setSelectedFiles(fileResult);
      });
    }
  }, [fileList]);

  return React.useMemo(() => ({
    files: selectedFiles,
  }), [selectedFiles]);
}

const FilePreviewGrid = ({
  fileList,
}: { fileList?: FileList }) => {
  const { files } = useFilePreviews(fileList);

  return (
    <div className="w-full h-full overflow-auto -mx-4 px-4">
      <div className="w-full grid gap-4 grid-cols-3">
        {files.map((f, i) => (
          <div
            data-testid="selectedFileItem"
            key={i}
            className={`w-full aspect-square rounded overflow-hidden relative before:absolute before:content-[''] before:bg-current before:top-0 before:left-0 before:w-full before:h-full before:opacity-10`}
            title={[f.name, getMimeTypeDescription(f.type), formatFileSize(f.size)].join(', ')}
          >
            {
              f.resolvedType === ContentType.IMAGE
              && typeof f.metadata?.previewUrl === 'string'
              && (
                <img
                  className="block w-full h-full object-center object-cover"
                  src={f.metadata.previewUrl}
                  alt={f.name}
                  data-testid="preview"
                />
              )
            }
            {
              f.resolvedType === ContentType.AUDIO
              && (
                <AudioMiniFilePreview file={f} />
              )
            }
          </div>
        ))}
      </div>
    </div>
  )
}

export const FileSelectBox = React.forwardRef<HTMLInputElement, FileButtonProps>(
  (
    {
      label = '',
      hint = '',
      border = false,
      block = false,
      enhanced = false,
      hiddenLabel = false,
      multiple = false,
      onChange,
      disabled = false,
      className: _className,
      placeholder: _placeholder,
      as: _as,
      ...etcProps
    }: FileButtonProps,
    forwardedRef,
  ) => {
    const [renderEnhanced, setRenderEnhanced] = React.useState(false);
    const [fileList, setFileList] = React.useState<FileList>();
    const [lastSelectedFileAt, setLastSelectedFileAt] = React.useState<number>();
    const defaultRef = React.useRef<HTMLInputElement>(null);
    const ref = forwardedRef ?? defaultRef;

    const addFile: React.ChangeEventHandler<HTMLInputElement> = (e) => {
      if (!enhanced) {
        onChange?.(e);
        return;
      }

      setFileList(e.currentTarget.files as FileList);
      setLastSelectedFileAt(Date.now());
      onChange?.(e);
    };

    const deleteFiles: React.MouseEventHandler<HTMLButtonElement> = () => {
      if (typeof ref === 'object' && ref.current) {
        ref.current.value = '';
        setFileList(undefined);
        setTimeout(() => {
          delegateTriggerChangeEvent(ref.current as HTMLInputElement);
        })
      }
    };

    const cancelEvent = (e: React.DragEvent) => {
      e.stopPropagation();
      e.preventDefault();
    }

    const handleDropZone: React.DragEventHandler<HTMLDivElement> = async (e) => {
      cancelEvent(e);
      const { dataTransfer } = e;
      if (typeof ref === 'object' && ref.current) {
        const { files } = dataTransfer;
        setFileList(ref.current.files = files);
        delegateTriggerChangeEvent(ref.current);
      }
    }

    React.useEffect(() => {
      setRenderEnhanced(enhanced);
    }, [enhanced]);

    return (
      <div
        className="block"
        onDragEnter={cancelEvent}
        onDragOver={cancelEvent}
        onDrop={handleDropZone}
        data-testid="root"
      >
        <label
          className="block absolute top-0 left-0 w-full h-full cursor-pointer"
          data-testid="clickArea"
        >
          <input
            {...etcProps}
            disabled={disabled}
            ref={ref}
            type="file"
            className={`${renderEnhanced ? 'sr-only' : ''}`}
            onChange={addFile}
            multiple={multiple}
            data-testid="input"
          />
        </label>
        {
          border && (
            <span
              data-testid="border"
              className="block"
            />
          )
        }
        {
          label
          && !hiddenLabel
          && (
            <div
              data-testid="label"
              className="block"
            >
              {label}
            </div>
          )
        }
        {
          (fileList?.length ?? 0) < 1
          && renderEnhanced
          && hint
          && (
            <div
              data-testid="hint"
              className="absolute top-0 left-0 w-full h-full pointer-events-none box-border overflow-hidden pt-4"
            >
              <div className="flex items-center justify-center w-full h-full">
                {hint}
              </div>
            </div>
          )
        }
        {
          (fileList?.length ?? 0) > 0
          && renderEnhanced
          && (
            <>
              <div className={`absolute top-0 left-0 w-full h-full pointer-events-none pb-12 box-border overflow-hidden pt-8`}>
                <div
                  className={`pointer-events-auto w-full h-full px-4 pb-4 box-border`}
                  key={lastSelectedFileAt}
                >
                  {
                    multiple
                    && (
                      <FilePreviewGrid fileList={fileList} />
                    )
                  }
                  {
                    !multiple
                    && (
                      <div
                        className={`w-full h-full`}
                      >
                        <div data-testid="selectedFileItem" className={`h-full w-full p-4 box-border rounded overflow-hidden relative before:absolute before:content-[''] before:bg-current before:top-0 before:left-0 before:w-full before:h-full before:opacity-10`}>
                          <FilePreviewComponent
                            fileList={fileList}
                          />
                        </div>
                      </div>
                    )
                  }
                </div>
              </div>
              <div className="pointer-events-none absolute bottom-0 left-0 w-full text-center h-12 box-border flex">
                <div className="w-0 flex-auto flex flex-col items-center justify-center h-full">
                  <span
                    className=""
                  >
                    Reselect
                  </span>
                </div>
                <div className="pointer-events-auto w-0 flex-auto flex flex-col items-center justify-center h-full">
                  <button
                    data-testid="clear"
                    type="button"
                    onClick={deleteFiles}
                    className=""
                  >
                    {multiple ? 'Clear' : 'Delete'}
                  </button>
                </div>
              </div>
            </>
          )
        }
      </div>
    );
  }
);

FileSelectBox.displayName = 'FileSelectBox';
